-- Based off RDX's alert system

local defaults = {
	profile = {
		-- Top
		TopScale = 0.8,
		TopGrowth = "DOWN",
		TopAlpha = 1,
		TopBarWidth = 250,
		-- Center
		CenterScale = 0.9,
		CenterGrowth = "UP",
		CenterAlpha = 1,
		CenterBarWidth = 275,
		-- Warning Bars
		WarningBars = true,
		WarningAnchor = true,
		WarningMessages = false,
		WarningScale = 1,
		WarningGrowth = "DOWN",
		WarningAlpha = 1,
		WarningBarWidth = 275,
		RedirectCenter = false,
		RedirectThreshold = 5,
		EnableWarningMessage = true,
		-- Warning Messages
		BeforeThreshold = 5,
		ClrWarningText = true,
		SinkStorage = {},
		SinkIcon = true,
		AnnounceToRaid = false,
		CdPopupMessage = false,
		CdBeforeMessage = true,
		DurPopupMessage = true,
		DurBeforeMessage = true,
		WarnPopupMessage = true,
		-- Flash
		FlashAlpha = 0.6,
		FlashDuration = 0.8,
		EnableOscillations = true,
		FlashOscillations = 2,
		FlashTexture = "Interface\\Tooltips\\UI-Tooltip-Background",
		ConstantClr = false,
		GlobalColor = {1,0,0},
		-- Text
		CenterTextWidth = 200,
		WarningTextWidth = 200,
		TopTextWidth = 180,
		TextXOffset = 10,
		TextYOffset = 1,
		TextAlpha = 1,
		-- Bar
		BarTextJustification = "LEFT",
		BarFontSize = 12,  --10,
		BarFillDirection = "FILL",
		--BarFontColor = {0.47, 0.75, 1, 1},
		--BarFontColor = {1,1,1,1},
		BarFontColor = {0.73, 0.73, 0.73, 1},
		BarHeight = 30,
		BarSpacing = 0,
		BarFontOutline = "OUTLINE", --THINOUTLINE
		AnimationSpeed = 0.3,
		-- Timer
		TimerFontColor = {1,1,1,1},
		TimerFontOutlineLeft = "DEFAULT",
		TimerFontOutlineRight = "DEFAULT",
		TimerXOffset = -5,
		TimerYOffset = 0,
		TimerAlpha = 1,
		TimerSecondsFontSize = 20,
		TimerDecimalFontSize = 12,
		ScaleTimerWithBarHeight = true,
		DecimalYOffset = 2,
		-- Icon
		IconXOffset = 0,
		IconYOffset = 0,
		IconSize = 30,
		SetIconToBarHeight = true,
		ShowLeftIcon = true,
		ShowRightIcon = false,
		-- Toggles
		DisableDropdowns = false,
		DisableScreenFlash = false,
		DisableSounds = false,
		ShowBarBorder = true,
		ShowIconBorder = true,
		MasterSound = true,
		DialogSound = false,
		SFXSound = false,
		MusicSound = false,
		AmbienceSound = false,
		NormalSound = false,
		-- Custom Bars
		CustomLocalClr = "TAN",
		CustomRaidClr = "TAN",
		CustomSound = "ALERT1",
		CustomAudioCD = false,
		CustomAudiotime = 5,
		CustomBiggerBar = false,		
		-- Pull Bars
		PullLocalClr = "RED",
		PullRaidClr = "RED",
		PullSound = "ALERT10",
		PullAudioCD = true,
		PullAudiotime = 5,
		PullBiggerBar = false,		
		-- Break Bars
		BreakLocalClr = "GREEN",
		BreakRaidClr = "GREEN",
		BreakSound = "ALERT14",	
		BreakAudioCD = true,
		BreakAudiotime = 5,
		BreakBiggerBar = false,
		-- Debuff Bars
		DebuffBars = true,
		DebuffAnchor = true,
		DebuffMessages = false,
		DebuffScale = 1,
		DebuffGrowth = "DOWN",
		DebuffAlpha = 1,
		DebuffBarWidth = 250,
		DebuffTextWidth = 250,
		-- Message Frame
		MessageBars = true,
		MessageAnchor = true,
		MessageMessages = false,
		MessageScale = 1,
		MessageGrowth = "UP",
		MessageAlpha = 0.75,
		MessageBarWidth = 250,
		MessageTextWidth = 250,
		-- Auras
		AurasScale = 0.8,
		AurasGrowth = "DOWN",
		AurasAlpha = 1,
		AurasBarWidth = 250,	
		AurasTextWidth = 180,
	}
}

local addon = DXE
local loader = DXE_Loader
local L = addon.L
local CN = addon.CN
local NID = addon.NID
local AceTimer = LibStub("AceTimer-3.0")

local Colors = addon.Media.Colors
local GetTime,PlaySoundFile = GetTime,PlaySoundFile
local ipairs,pairs,next,remove = ipairs,pairs,next,table.remove
local util = addon.util
local find = string.find

local FADE_TIME = 1
local INSET = 2
local ANNOUNCE_FORMAT = "*** %s ***"

local db,pfl,gbl
local ANIMATION_TIME -- 0.3

---------------------------------------
-- INITIALIZATION
---------------------------------------

local module = addon:NewModule("Alerts","LibSink-2.0","AceEvent-3.0","AceTimer-3.0","AceComm-3.0","AceSerializer-3.0")
addon.Alerts = module
local Active = {}
local TopAlertStack = {}
local CenterAlertStack = {}
local WarningAlertStack = {}
local DebuffAlertStack = {}
local MessageAlertStack = {}
local AurasAlertStack = {}
local BarPool = {}
local MessagePool = {}
local prototype = {}

local TopStackAnchor,CenterStackAnchor,WarningStackAnchor, DebuffStackAnchor, MessageStackAnchor, AurasAnchor

function module:RefreshProfile()
	pfl = db.profile
	self:RefreshBars()
	self:SetSinkStorage(pfl.SinkStorage)
	self:UpdateFlashSettings()
end

function module:OnInitialize()
	-- Top stack anchor
	TopStackAnchor = addon:CreateLockableFrame("AlertsTopStackAnchor",245,10,format("%s - %s",L["Alerts"],L["Bar Top Anchor"]))
	addon:RegisterMoveSaving(TopStackAnchor,"RIGHT","UIParent","RIGHT",-330.5000126174457,23.49995633474761)
	--addon:RegisterMoveSaving(TopStackAnchor,"CENTER","UIParent","CENTER",480.5000126174457,23.49995633474761)
	addon:LoadPosition("DXEAlertsTopStackAnchor")

	-- Bottom stack anchor
	CenterStackAnchor = addon:CreateLockableFrame("AlertsCenterStackAnchor",245,10,format("%s - %s",L["Alerts"],L["Bar Center Anchor"]))
	addon:RegisterMoveSaving(CenterStackAnchor,"CENTER","UIParent","CENTER",0,-235.9996233230851)
	addon:LoadPosition("DXEAlertsCenterStackAnchor")
	
	-- Warning anchor
	WarningStackAnchor = addon:CreateLockableFrame("AlertsWarningStackAnchor",245,10,format("%s - %s",L["Alerts"],L["Bar Warning Anchor"]))
	addon:RegisterMoveSaving(WarningStackAnchor,"CENTER","UIParent","CENTER",0,246.9998803907182)
	addon:LoadPosition("DXEAlertsWarningStackAnchor")

	-- Inform anchor
	InformStackAnchor = addon:CreateLockableFrame("AlertsInformStackAnchor",245,10,format("%s - %s",L["Alerts"],L["Bar Inform Anchor"]))
	addon:RegisterMoveSaving(InformStackAnchor,"CENTER","UIParent","CENTER",0,-60)
	addon:LoadPosition("DXEAlertsInformStackAnchor")
	
	DebuffStackAnchor = addon:CreateLockableFrame("AlertsDebuffStackAnchor",245,10,format("%s - %s",L["Alerts"],L["Raid Debuff Anchor"]))
	addon:RegisterMoveSaving(DebuffStackAnchor,"LEFT","UIParent","LEFT",312.4999134559484,20)
	--addon:RegisterMoveSaving(DebuffStackAnchor,"CENTER","UIParent","CENTER",-480.4999134559484,20)
	addon:LoadPosition("DXEAlertsDebuffStackAnchor")
	
	MessageStackAnchor = addon:CreateLockableFrame("AlertsMessageStackAnchor",245,10,format("%s - %s",L["Alerts"],L["Message Anchor"]))
	addon:RegisterMoveSaving(MessageStackAnchor,"CENTER","UIParent","CENTER",0,21.00003180006635)
	addon:LoadPosition("DXEAlertsMessageStackAnchor")
	
	-- Auras anchor
	AurasAnchor = addon:CreateLockableFrame("AlertsAurasAnchor",245,10,format("%s - %s",L["Alerts"],L["Auras Anchor"]))
	addon:RegisterMoveSaving(AurasAnchor,"LEFT","UIParent","LEFT",312.4999134559484,246.9998803907182)
	--addon:RegisterMoveSaving(AurasAnchor,"CENTER","UIParent","CENTER",-480.4999134559484,246.9998803907182)
	addon:LoadPosition("DXEAlertsAurasAnchor")
	
	self.db = addon.db:RegisterNamespace("Alerts", defaults)
	db = self.db
	pfl = db.profile
	gbl = addon.db

	self:SetSinkStorage(pfl.SinkStorage)
	self:UpdateFlashSettings()

	self:RegisterComm("D4")
	self:RegisterComm("BigWigs")
	
	db.RegisterCallback(self, "OnProfileChanged", "RefreshProfile")
	db.RegisterCallback(self, "OnProfileCopied", "RefreshProfile")
	db.RegisterCallback(self, "OnProfileReset", "RefreshProfile")
	
	module:RegisterEvent("GROUP_ROSTER_UPDATE")
	module:GROUP_ROSTER_UPDATE()
	
	local tbl = DXEDB.breakTime
	if tbl then -- Break time present, resume it
		local prevTime, seconds, text = tbl[1], tbl[2], tbl[3], tbl[4]
		local curTime = time()
		if curTime-prevTime > seconds then
			DXEDB.breakTime = nil
		else
			module:BreakRelog(seconds-(curTime-prevTime), text)
		end
	end
		
	ANIMATION_TIME = gbl.profile.Globals.AnimationSpeed
	
	-- To fix broken configurations in v668, to be removed
	if pfl.TimerFontOutlineLeft == "NORMAL" then pfl.TimerFontOutlineLeft = "DEFAULT" end
	if pfl.TimerFontOutlineRight == "NORMAL" then pfl.TimerFontOutlineLeft = "DEFAULT" end
	if pfl.BarFontOutline == "NORMAL" then pfl.BarFontOutline = "DEFAULT" end
end

--function module:OnEnable()
--	self:RegisterComm("D4")
--	self:RegisterComm("BigWigs")
--end

function module:OnDisable()
	self:QuashAll()
end

---------------------------------------
-- UTILITY
---------------------------------------

local SM = LibStub("LibSharedMedia-3.0")
local Sounds = addon.Media.Sounds
local floor = math.floor
local ceil = math.ceil
local gsub = string.gsub

local function GetMedia(sound,c1,c2)
	if c1 == "GLOBAL" then
		local soundFile = Sounds:GetFile(sound)
		local c1Data = Colors[gbl.profile.Globals.GlobalBarCrl]
		local c2Data = Colors[c2]
		return soundFile,c1Data,c2Data
	else 
		return Sounds:GetFile(sound),Colors[c1],Colors[c2] 
	end
end
function module:GetMediaT(sound,c1,c2)
	return GetMedia(sound,c1,c2)
end

local function GetMessageEra(id)
	local popup,before

	if find(id,"cd$") then
		popup = pfl.CdPopupMessage
		before = pfl.CdBeforeMessage
	elseif find(id,"dur$") then
		popup = pfl.DurPopupMessage
		before = pfl.DurBeforeMessage
	elseif find(id,"self$") or find(id,"you$") then
		popup = pfl.DurPopupMessage
	elseif find(id,"warn$") then
		popup = pfl.WarnPopupMessage
	end

	return popup,before
end

local function MMSS(time)
	local min = floor(time/60)
	local sec = ceil(time % 60)
	return ("%d:%02d"):format(min,sec)
end

local ColorText

do
	local function colorlist(word,comma)
		return CN[word]..comma
	end

	local function helper(prefix,word)
		-- comma separated
		if find(word,"^[^ ,]+,") then
			return prefix..gsub(word,"([^ ,]+)(,?)",colorlist)
		else
			return prefix..CN[word]
		end
	end

	function ColorText(text)
		return (gsub(text,"^(.+: )([^!.]+)",helper))
	end
end

local function Pour(text,icon,color)
	color = color or Colors.WHITE
	if not pfl.SinkIcon then icon = nil end
	local pour_text = pfl.ClrWarningText and ColorText(text) or text
	module:Pour(pour_text,color.r,color.g,color.b,nil,nil,nil,nil,nil,icon)
	
	local no_you = not find(text,L.alert["YOU"])
	if no_you and pfl.AnnounceToRaid and addon:IsPromoted() and addon.GroupType == "RAID" then
		-- color coded text disconnects the user!!!
		text = gsub(text,"|c%x%x%x%x%x%x%x%x(.-)|r","%1")
		SendChatMessage(format(ANNOUNCE_FORMAT,text),"RAID_WARNING")
	end
	--if icon then text = "|T"..icon..":20:20:-5|t"..text end
	--UIErrorsFrame:AddMessage(text, r, g, b, 1, UIERRORS_HOLD_TIME)
end

local function Sound(...)
	if pfl.MasterSound then
		PlaySoundFile(...,"Master")
	elseif pfl.DialogSound then
		PlaySoundFile(...,"Dialog")	
	elseif pfl.SFXSound then
		PlaySoundFile(...,"SFX")		
	elseif pfl.MusicSound then
		PlaySoundFile(...,"Music")	
	elseif pfl.AmbienceSound then
		PlaySoundFile(...,"Ambience")			
	else
		PlaySoundFile(...)
	end
end
function module:SoundT(...)
	Sound(...)
end
---------------------------------------------
-- SCREEN FLASH
---------------------------------------------

do
	local FLASH_DURATION,PERIOD,AMP,MULT,OSC

	local flash = CreateFrame("Frame","DXEAlertsFlash",UIParent)
	flash:SetFrameStrata("BACKGROUND")
	local t = flash:CreateTexture(nil,"BACKGROUND")
	t:SetAllPoints(true)
	flash:SetAllPoints(true)
	flash:Hide()

	local counter
	local function OnUpdate(self,elapsed)
		counter = counter + elapsed
		if counter > FLASH_DURATION then self:Hide() end
		if OSC then
			local p = counter % PERIOD
			if p > AMP then p = PERIOD - p end
			self:SetAlpha(p * MULT)
		end
	end

	flash:SetScript("OnUpdate",OnUpdate)

	module.FlashTextures = {
		["Interface\\Tooltips\\UI-Tooltip-Background"] = L["Solid"],
		["Interface\\Addons\\DXE\\Textures\\LowHealthGray"] = L["Low Health"],
	}

	function module:UpdateFlashSettings()
		t:SetTexture(pfl.FlashTexture)
	end

	function module:FlashScreen(c)
		if pfl.DisableScreenFlash then return end
		if pfl.ConstantClr then
			local r,g,b = unpack(pfl.GlobalColor)
			t:SetVertexColor(r,g,b,pfl.FlashAlpha)
		else
			c = c or Colors.BLACK
			t:SetVertexColor(c.r,c.g,c.b,pfl.FlashAlpha)
		end
		counter = 0
		OSC = pfl.EnableOscillations
		FLASH_DURATION = pfl.FlashDuration
		PERIOD = FLASH_DURATION / pfl.FlashOscillations
		AMP = PERIOD / 2
		MULT = 1 / AMP
		flash:SetAlpha(OSC and 0 or 1)
		flash:Show()
	end
end

---------------------------------------
-- UPDATING
---------------------------------------

local function OnUpdate(self,elapsed)
	local bar = next(Active)
	if not bar then self:Hide() return end
	local time = GetTime()
	while bar do
		if bar.countFunc then bar:countFunc(time) end
		if bar.animFunc then bar:animFunc(time) end
		bar = next(Active,bar)
	end
end

local UpdateFrame = CreateFrame("Frame",nil,UIParent)
UpdateFrame:SetScript("OnUpdate",OnUpdate)
UpdateFrame:Hide()

---------------------------------------
-- PROTOTYPE
---------------------------------------
local wipe = table.wipe

function prototype:SetColor(c1,c2)
	self.data.c1 = c1
	self.statusbar:SetStatusBarColor(c1.r,c1.g,c1.b)
	self.data.c2 = c2 or c1
end

function prototype:Destroy()
	self:Hide()
	self:ClearAllPoints()
	self:UnregisterAllEvents()
	self:RemoveFromStacks()
	self:CancelAllTimers()
	self.countFunc = nil
	self.animFunc = nil
	self.statusbar:SetValue(0)
	Active[self] = nil
	UIFrameFadeRemoveFrame(self)
	BarPool[self] = true
	wipe(self.data)
	self.timer:Show()
	self.lefticon:Hide()
	self.lefticon.t:SetTexture("")
	self.righticon:Hide()
	self.righticon.t:SetTexture("")
	self.bmsg = nil
	self.orig_text = nil
	self:CancelAudioCountdown()
	self:LayoutAlertStack(TopAlertStack, TopStackAnchor, pfl.TopGrowth)
	self:LayoutAlertStack(CenterAlertStack, CenterStackAnchor, pfl.CenterGrowth)
	self:LayoutAlertStack(WarningAlertStack, WarningStackAnchor, pfl.WarningGrowth)
	self:LayoutAlertStack(DebuffAlertStack, DebuffStackAnchor, pfl.DebuffGrowth)
	self:LayoutAlertStack(MessageAlertStack, MessageStackAnchor, pfl.MessageGrowth)
	self:LayoutAlertStack(AurasAlertStack, AurasAnchor, pfl.AurasGrowth)
end

do
	local function SortDesc(a1, a2)
		return (a1.data.timeleft or 10000) < (a2.data.timeleft or 10000)
	end

	function prototype:LayoutAlertStack(stack,anchor,growth)
		sort(stack, SortDesc)
		local point,relpoint,mult
		if growth == "DOWN" then
			point,relpoint,mult = "TOP","BOTTOM",-1
		elseif growth == "UP" then
			point,relpoint,mult = "BOTTOM","TOP",1
		end
		for i,bar in ipairs(stack) do
			bar:ClearAllPoints()
			bar:SetPoint(point,anchor,relpoint,0,mult*(i-1)*(pfl.BarHeight + pfl.BarSpacing))
		end
	end
end

function prototype:AnchorToAuras()
	self.data.anchor = "AURAS"
	self:SetAlpha(pfl.AurasAlpha)
	self:SetScale(pfl.AurasScale)
	self:SetWidth(pfl.AurasAlpha)

	self.text:SetWidth(pfl.AurasTextWidth)
	self:RemoveFromStacks()
	AurasAlertStack[#AurasAlertStack+1] = self
	self:LayoutAlertStack(AurasAlertStack, AurasAnchor, pfl.AurasGrowth)
end

function prototype:AnchorToTop()
	self.data.anchor = "TOP"
	self:SetAlpha(pfl.TopAlpha)
	self:SetScale(pfl.TopScale)
	self:SetWidth(pfl.TopBarWidth)

	self.text:SetWidth(pfl.TopTextWidth)
	self:RemoveFromStacks()
	TopAlertStack[#TopAlertStack+1] = self
	self:LayoutAlertStack(TopAlertStack, TopStackAnchor, pfl.TopGrowth)
end

function prototype:AnchorToCenter()
	if self.data.sound and not pfl.DisableSounds then Sound(self.data.sound) end
	self.data.anchor = "CENTER"
	self:SetAlpha(pfl.CenterAlpha)
	self:SetScale(pfl.CenterScale)
	self:SetWidth(pfl.CenterBarWidth)
	self.text:SetWidth(pfl.CenterTextWidth)
	self:RemoveFromStacks()
	CenterAlertStack[#CenterAlertStack+1] = self
	self:LayoutAlertStack(CenterAlertStack, CenterStackAnchor, pfl.CenterGrowth)
end

function prototype:AnchorToWarning()
	if self.data.sound and not pfl.DisableSounds then Sound(self.data.sound) end
	self.data.anchor = "WARNING"
	self:SetAlpha(pfl.WarningAlpha)
	self:SetScale(pfl.WarningScale)
	self:SetWidth(pfl.WarningBarWidth)
	self.text:SetWidth(pfl.WarningTextWidth)
	self:RemoveFromStacks()
	WarningAlertStack[#WarningAlertStack+1] = self
	self:LayoutAlertStack(WarningAlertStack, WarningStackAnchor, pfl.WarningGrowth)
end

function prototype:AnchorToDebuff()
	if self.data.sound and not pfl.DisableSounds then Sound(self.data.sound) end
	self.data.anchor = "DEBUFF"
	self:SetAlpha(pfl.DebuffAlpha)
	self:SetScale(pfl.DebuffScale)
	self:SetWidth(pfl.DebuffBarWidth)
	self.text:SetWidth(pfl.DebuffTextWidth)
	self:RemoveFromStacks()
	DebuffAlertStack[#DebuffAlertStack+1] = self
	self:LayoutAlertStack(DebuffAlertStack, DebuffStackAnchor, pfl.DebuffGrowth)
end
function prototype:AnchorToMessage()
	if self.data.sound and not pfl.DisableSounds then Sound(self.data.sound) end
	self.data.anchor = "MESSAGE"
	self:SetAlpha(pfl.MessageAlpha)
	self:SetScale(pfl.MessageScale)
	self:SetWidth(pfl.MessageBarWidth)
	self.text:SetWidth(pfl.MessageTextWidth)
	self:RemoveFromStacks()
	MessageAlertStack[#MessageAlertStack+1] = self
	self:LayoutAlertStack(MessageAlertStack, MessageStackAnchor, pfl.MessageGrowth)
end
do
	local UIParent = UIParent
	local function AnimationFunc(self,time)
		local data = self.data
		local perc = (time - data.t0) / ANIMATION_TIME
		if perc < 0 or perc > 1 then
			self.animFunc = nil
			self:AnchorToCenter()
			if self.data.flashscreen then
				module:FlashScreen(self.data.c1)
			end
		else
			local a = pfl.TopAlpha + ((pfl.CenterAlpha - pfl.TopAlpha) * perc)
			self:SetAlpha(a)

			local s = pfl.TopScale + ((pfl.CenterScale - pfl.TopScale) * perc)
			self:SetScale(s)

			local w = pfl.TopBarWidth + ((pfl.CenterBarWidth - pfl.TopBarWidth) * perc)
			self:SetWidth(w)

			local escale = self:GetEffectiveScale()
			local x = (data.fx + ((data.tox - data.fx) * perc)) / escale
			local y = (data.fy + ((data.toy - data.fy) * perc)) / escale
			self:ClearAllPoints()
			self:SetPoint("CENTER",UIParent,"BOTTOMLEFT",x,y)
		end
	end

	function prototype:TranslateToCenter()
		if self.paused == true then
			--print("TranslateToCenterpaused",x,y,self:GetID())
			return
		end	
		self.data.anchor = nil
		self:RemoveFromStacks()
		local x,y = self:GetCenter()

		if not x then
			--print("TranslateToCenter",x,y,self:GetID())
			return false
		end
		
		self:ClearAllPoints()
		self:SetPoint("CENTER",UIParent,"BOTTOMLEFT",x,y)
		self:LayoutAlertStack(TopAlertStack, TopStackAnchor, pfl.TopGrowth)

		local worldscale,escale = UIParent:GetEffectiveScale(),self:GetEffectiveScale()
		local fx,fy = x*escale, y*escale
		local cx,cy = CenterStackAnchor:GetCenter()
		local mult = pfl.CenterGrowth == "DOWN" and -1 or 1
		cy = cy + mult*5 -- CenterStackAnchor:GetHeight() / 2

		local offset = (pfl.BarHeight * #CenterAlertStack + pfl.BarHeight/2) * (pfl.CenterScale * worldscale)
		local tox,toy = cx*worldscale,cy*worldscale
		toy = toy + mult*(offset)

		local data = self.data
		data.t0 = GetTime()
		data.fx = fx
		data.fy = fy
		data.tox = tox
		data.toy = toy
		self.animFunc = AnimationFunc
		self:LayoutAlertStack(CenterAlertStack, CenterStackAnchor, pfl.CenterGrowth)
	end
end

function prototype:RemoveFromStack(stack)
	for i,bar in ipairs(stack) do
		if bar == self then
			remove(stack,i)
			return
		end
	end
end

function prototype:RemoveFromStacks()
	self:RemoveFromStack(TopAlertStack)
	self:RemoveFromStack(CenterAlertStack)
	self:RemoveFromStack(WarningAlertStack)
	self:RemoveFromStack(DebuffAlertStack)
	self:RemoveFromStack(AurasAlertStack)
end

do
	local function CountdownFunc(self,time)
		local timeleft = self.data.endTime - time
		local data = self.data
		data.timeleft = timeleft
		if timeleft < 0 then
			self.countFunc = nil
			self.timer:SetTime(0)
			return
		end
		self.timer:SetTime(timeleft)
		local value = 1 - (timeleft / data.totalTime)
		self.statusbar:SetValue(pfl.BarFillDirection == "FILL" and value or 1 - value)

		if pfl.WarningMessages and self.bmsg and timeleft <= pfl.BeforeThreshold then
			self.bmsg = nil
			Pour(self.orig_text.." - "..MMSS(timeleft),data.icon,data.c1)
		end
	end

	local cos = math.cos
	local function CountdownFlashFunc(self,time)
		if self.paused == true then
			--print("CountdownFlashFunc",bar:GetID(),bar.paused)
			return
		end		
		
		local data = self.data
		local timeleft = data.endTime - time
		self.data.timeleft = timeleft

		if timeleft < 0 then
			self.countFunc = nil
			self.timer:SetTime(0)
			return
		end
		self.timer:SetTime(timeleft)

		local value = 1 - (timeleft / data.totalTime)
		self.statusbar:SetValue(pfl.BarFillDirection == "FILL" and value or 1 - value)
		if timeleft < data.flashTime then
			self.statusbar:SetStatusBarColor(util.blend(data.c1, data.c2, 0.5*(cos(timeleft*12) + 1)))
		end

		if pfl.WarningMessages and self.bmsg and timeleft <= pfl.BeforeThreshold then
			self.bmsg = nil
			Pour(self.orig_text.." - "..MMSS(timeleft),data.icon,data.c1)
		end
	end

	-- for absorb bar
	local function CountdownNoSetFlashFunc(self,time)
		local data = self.data
		local timeleft = data.endTime - time
		self.data.timeleft = timeleft
		if timeleft < 0 then
			self.countFunc = nil
			self.timer:SetTime(0)
			return
		end
		self.timer:SetTime(timeleft)
		if timeleft < data.flashTime then
			self.statusbar:SetStatusBarColor(util.blend(data.c1, data.c2, 0.5*(cos(timeleft*12) + 1)))
		end
	end

	function prototype:Countdown(totalTime, flashTime, noset)
		local endTime = GetTime() + totalTime
		self.data.endTime,self.data.totalTime = endTime, totalTime
		if noset then
			self.data.flashTime = flashTime
			self.countFunc = CountdownNoSetFlashFunc
		elseif flashTime and self.data.c1 ~= self.data.c2 then
			self.data.flashTime = flashTime
			self.countFunc = CountdownFlashFunc
		else
			self.countFunc = CountdownFunc
		end
	end
end

function prototype:Fade()
	if self.paused == true then
		return
	end
	local fadeTable = self.fadeTable
	fadeTable.startAlpha = self:GetAlpha()
	fadeTable.fadeTimer = 0
	fadeTable.finishedFunc = self.Destroy
	UIFrameFade(self,fadeTable)
end

function prototype:SetID(id)
	self.data.id = id
end

function prototype:GetID(id)
	return self.data.id
end

function prototype:SetTimeleft(timeleft)
	self.data.timeleft = timeleft
end

function prototype:SetSound(sound)
	self.data.sound = sound
end

function prototype:SetFlashScreen(flashscreen)
	self.data.flashscreen = flashscreen
end

function prototype:SetText(text)
	self.text:SetText(ColorText(text))
	self.orig_text = text
end
local role_tex_file = "Interface\\LFGFrame\\UI-LFG-ICON-PORTRAITROLES.blp"
local role_t = "\124T"..role_tex_file..":%d:%d:"
local role_tex = {
   DAMAGER = role_t.."0:0:64:64:20:39:22:41\124t",
   HEALER  = role_t.."0:0:64:64:20:39:1:20\124t",
   TANK    = role_t.."0:0:64:64:0:19:22:41\124t",
   LEADER  = role_t.."0:0:64:64:0:19:1:20\124t",
}
local function getRoleTexCoord(role)
  local str = role_tex[role]
  if not str or #str == 0 then return nil end
  local a,b,c,d = string.match(str, ":(%d+):(%d+):(%d+):(%d+)%\124t")
  return a/64,b/64,c/64,d/64
end
function prototype:SetIcon(texture)
	if not texture then
		self.lefticon:Hide()
		self.righticon:Hide()
		return
	end

	self.data.icon = texture

	if pfl.ShowLeftIcon then
		self.lefticon:Show()
		if texture == "TANK_ICON" then
			self.lefticon.t:SetTexture(role_tex_file)
			self.lefticon.t:SetTexCoord(getRoleTexCoord("TANK"))
		elseif texture == "HEALER_ICON" then
			self.lefticon.t:SetTexture(role_tex_file)
			self.lefticon.t:SetTexCoord(getRoleTexCoord("HEALER"))
		elseif texture == "DAMAGE_ICON" then
			self.lefticon.t:SetTexture(role_tex_file)
			self.lefticon.t:SetTexCoord(getRoleTexCoord("DAMAGER"))			
		else
			self.lefticon.t:SetTexture(texture)
		end
	end

	if pfl.ShowRightIcon then
		self.righticon:Show()
		if texture == "TANK_ICON" then
			self.lefticon.t:SetTexture(role_tex_file)
			self.lefticon.t:SetTexCoord(getRoleTexCoord("TANK"))
		elseif texture == "HEALER_ICON" then
			self.lefticon.t:SetTexture(role_tex_file)
			self.lefticon.t:SetTexCoord(getRoleTexCoord("HEALER"))
		elseif texture == "DAMAGE_ICON" then
			self.lefticon.t:SetTexture(role_tex_file)
			self.lefticon.t:SetTexCoord(getRoleTexCoord("DAMAGER"))			
		else
			self.lefticon.t:SetTexture(texture)
		end
	end
end

function prototype:FireBeforeMsg() self.bmsg = true end

local function SkinBar(bar)
	if not pfl.ShowLeftIcon then bar.lefticon:Hide() end
	if not pfl.ShowRightIcon then bar.righticon:Hide() end
	bar:SetIcon(bar.data.icon)

	bar.statusbar:ClearAllPoints()
	bar.lefticon.t:ClearAllPoints()
	bar.righticon.t:ClearAllPoints()
	bar.bg:ClearAllPoints()

	if pfl.ShowBarBorder then
		bar.border:Show()
		bar.bg:SetPoint("TOPLEFT",INSET,-INSET)
		bar.bg:SetPoint("BOTTOMRIGHT",-INSET,INSET)
		bar.statusbar:SetPoint("TOPLEFT",INSET,-INSET)
		bar.statusbar:SetPoint("BOTTOMRIGHT",-INSET,INSET)
	else
		bar.border:Hide()
		bar.statusbar:SetPoint("TOPLEFT")
		bar.statusbar:SetPoint("BOTTOMRIGHT")
		bar.bg:SetPoint("TOPLEFT")
		bar.bg:SetPoint("BOTTOMRIGHT")
	end

	if pfl.ShowIconBorder then
		bar.lefticon.border:Show()
		bar.lefticon.t:SetPoint("TOPLEFT",bar.lefticon,"TOPLEFT",INSET,-INSET)
		bar.lefticon.t:SetPoint("BOTTOMRIGHT",bar.lefticon,"BOTTOMRIGHT",-INSET,INSET)

		bar.righticon.border:Show()
		bar.righticon.t:SetPoint("TOPLEFT",bar.righticon,"TOPLEFT",INSET,-INSET)
		bar.righticon.t:SetPoint("BOTTOMRIGHT",bar.righticon,"BOTTOMRIGHT",-INSET,INSET)
	else
		bar.righticon.border:Hide()
		bar.righticon.t:SetPoint("TOPLEFT",bar.righticon,"TOPLEFT")
		bar.righticon.t:SetPoint("BOTTOMRIGHT",bar.righticon,"BOTTOMRIGHT")

		bar.lefticon.border:Hide()
		bar.lefticon.t:SetPoint("TOPLEFT",bar.lefticon,"TOPLEFT")
		bar.lefticon.t:SetPoint("BOTTOMRIGHT",bar.lefticon,"BOTTOMRIGHT")
	end

	bar:SetHeight(pfl.BarHeight)

	if pfl.ScaleTimerWithBarHeight then
		if pfl.TimerFontOutlineLeft ~= "DEFAULT" then
			bar.timer.left:SetFont((bar.timer.left:GetFont()),(0.4375*pfl.BarHeight)+6.875, pfl.TimerFontOutlineLeft)
		else
			bar.timer.left:SetFont((bar.timer.left:GetFont()),(0.4375*pfl.BarHeight)+6.875)
		end
		if pfl.TimerFontOutlineRight ~= "DEFAULT" then
			bar.timer.right:SetFont((bar.timer.right:GetFont()),(0.25*pfl.BarHeight)+3.1, pfl.TimerFontOutlineRight)
		else
			bar.timer.right:SetFont((bar.timer.right:GetFont()),(0.25*pfl.BarHeight)+4.5)
		end
	else
		if pfl.TimerFontOutlineLeft ~= "DEFAULT" then
			bar.timer.left:SetFont((bar.timer.left:GetFont()),pfl.TimerSecondsFontSize, pfl.TimerFontOutlineLeft)
		else
			bar.timer.left:SetFont((bar.timer.left:GetFont()),pfl.TimerSecondsFontSize)
		end
		if pfl.TimerFontOutlineRight ~= "DEFAULT" then
			bar.timer.right:SetFont((bar.timer.right:GetFont()),pfl.TimerDecimalFontSize, pfl.TimerFontOutlineRight)
		else
			bar.timer.right:SetFont((bar.timer.right:GetFont()),pfl.TimerDecimalFontSize)
		end
	end	
	bar.timer.right:ClearAllPoints()
	bar.timer.right:SetPoint("BOTTOMLEFT",bar.timer.left,"BOTTOMRIGHT",0,pfl.DecimalYOffset)

	bar.timer:ClearAllPoints()
	bar.timer:SetPoint("RIGHT",bar,"RIGHT",pfl.TimerXOffset,pfl.TimerYOffset)
	bar.timer.right:SetVertexColor(unpack(pfl.TimerFontColor))
	bar.timer.left:SetVertexColor(unpack(pfl.TimerFontColor))
	bar.timer:SetAlpha(pfl.TimerAlpha)

	bar.lefticon:ClearAllPoints()
	bar.lefticon:SetPoint("RIGHT",bar,"LEFT",-pfl.IconXOffset,pfl.IconYOffset)

	bar.righticon:ClearAllPoints()
	bar.righticon:SetPoint("LEFT",bar,"RIGHT",pfl.IconXOffset,pfl.IconYOffset)

	if pfl.SetIconToBarHeight then
		bar.lefticon:SetWidth(pfl.BarHeight)
		bar.lefticon:SetHeight(pfl.BarHeight)
		bar.righticon:SetWidth(pfl.BarHeight)
		bar.righticon:SetHeight(pfl.BarHeight)
	else
		bar.lefticon:SetWidth(pfl.IconSize)
		bar.lefticon:SetHeight(pfl.IconSize)
		bar.righticon:SetWidth(pfl.IconSize)
		bar.righticon:SetHeight(pfl.IconSize)
	end
	if not pfl.BarFontOutline then pfl.BarFontOutline = "OUTLINE" end
	if not pfl.BarFontSize then pfl.BarFontSize = 12 end
	if pfl.BarFontOutline ~= "DEFAULT" then
		bar.text:SetFont(bar.text:GetFont(),pfl.BarFontSize, pfl.BarFontOutline)
	else
		bar.text:SetFont(bar.text:GetFont(),pfl.BarFontSize)
	end
	bar.text:SetHeight(pfl.BarFontSize * 1.2)
	bar.text:SetVertexColor(unpack(pfl.BarFontColor))
	bar.text:SetJustifyH(pfl.BarTextJustification)
	bar.text:SetPoint("LEFT",bar,"LEFT",pfl.TextXOffset,pfl.TextYOffset)
	bar.text:SetAlpha(pfl.TextAlpha)

	local data = bar.data
	if data.anchor == "TOP" then
		bar:SetScale(pfl.TopScale)
		bar:SetAlpha(pfl.TopAlpha)
		bar:SetWidth(pfl.TopBarWidth)
		bar.text:SetWidth(pfl.TopTextWidth)
	elseif data.anchor == "CENTER" then
		bar:SetScale(pfl.CenterScale)
		bar:SetAlpha(pfl.CenterAlpha)
		bar:SetWidth(pfl.CenterBarWidth)
		bar.text:SetWidth(pfl.CenterTextWidth)
	elseif data.anchor == "WARNING" then
		bar:SetScale(pfl.WarningScale)
		bar:SetAlpha(pfl.WarningAlpha)
		bar:SetWidth(pfl.WarningBarWidth)
		bar.text:SetWidth(pfl.WarningTextWidth)
	elseif data.anchor == "DEBUFF" then
		bar:SetScale(pfl.DebuffScale)
		bar:SetAlpha(pfl.DebuffAlpha)
		bar:SetWidth(pfl.DebuffBarWidth)
		bar.text:SetWidth(pfl.DebuffTextWidth)
	elseif data.anchor == "AURAS" then
		bar:SetScale(pfl.AurasScale)
		bar:SetAlpha(pfl.AurasAlpha)
		bar:SetWidth(pfl.AurasBarWidth)
		bar.text:SetWidth(pfl.AurasTextWidth)		
	end
	
end

function module:RefreshBars()
	ANIMATION_TIME = gbl.profile.Globals.AnimationSpeed
	if not next(Active) and not next(BarPool) then return end
	for bar in pairs(Active) do SkinBar(bar) end
	for bar in pairs(BarPool) do SkinBar(bar) end
	prototype:LayoutAlertStack(TopAlertStack, TopStackAnchor, pfl.TopGrowth)
	prototype:LayoutAlertStack(CenterAlertStack, CenterStackAnchor, pfl.CenterGrowth)
	prototype:LayoutAlertStack(WarningAlertStack, WarningStackAnchor, pfl.WarningGrowth)
	prototype:LayoutAlertStack(DebuffAlertStack, DebuffStackAnchor, pfl.DebuffGrowth)
	prototype:LayoutAlertStack(MessageAlertStack, MessageStackAnchor, pfl.MessageGrowth)
	prototype:LayoutAlertStack(AurasAlertStack, AurasAnchor, pfl.AurasGrowth)
end

local BarCount = 1
local function CreateBar()
	local self = CreateFrame("Frame","DXEAlertBar"..BarCount,UIParent)
	self:SetHeight(pfl.BarHeight)

	local bg = self:CreateTexture(nil,"BACKGROUND")
	addon:RegisterBackground(bg)
	self.bg = bg

	self.data = {}

	local statusbar = CreateFrame("StatusBar",nil,self)
	statusbar:SetMinMaxValues(0,1)
	statusbar:SetValue(0)
	addon:RegisterStatusBar(statusbar)
	self.statusbar = statusbar

	local border = CreateFrame("Frame",nil,self)
	border:SetAllPoints(true)
	border:SetFrameLevel(statusbar:GetFrameLevel()+1)
	addon:RegisterBorder(border)
	self.border = border

	local timer = addon.Timer:New(self)
	timer:SetFrameLevel(self:GetFrameLevel()+1)
	timer.left:SetShadowOffset(1,-1)
	timer.right:SetShadowOffset(1,-1)
	self.timer = timer

	local text = statusbar:CreateFontString(nil,"ARTWORK")
	text:SetShadowOffset(1,-1)
	text:SetNonSpaceWrap(true)
	addon:RegisterFontString(text,10)
	self.text = text

	local lefticon = CreateFrame("Frame",nil,self)
	self.lefticon = lefticon
	lefticon.t = lefticon:CreateTexture(nil,"BACKGROUND")
	lefticon.t:SetTexCoord(0.07,0.93,0.07,0.93)

	lefticon.border = CreateFrame("Frame",nil,lefticon)
	lefticon.border:SetAllPoints(true)
	addon:RegisterBorder(lefticon.border)

	local righticon = CreateFrame("Frame",nil,self)
	self.righticon = righticon
	righticon.t = righticon:CreateTexture(nil,"BACKGROUND")
	righticon.t:SetTexCoord(0.07,0.93,0.07,0.93)

	righticon.border = CreateFrame("Frame",nil,righticon)
	righticon.border:SetAllPoints(true)
	addon:RegisterBorder(righticon.border)

	AceTimer:Embed(self)
	for k,v in pairs(prototype) do self[k] = v end

	self.fadeTable = {mode = "OUT", timeToFade = FADE_TIME, endAlpha = 0, finishedArg1 = self }

	BarCount = BarCount + 1

	SkinBar(self)

	return self
end

local function GetBar()
	local bar = next(BarPool)
	if bar then BarPool[bar] = nil
	else bar = CreateBar() end
	Active[bar] = true
	UpdateFrame:Show()
	bar:Show()

	return bar
end

---------------------------------------
-- AUDIO COUNTDOWN
---------------------------------------
local CDList = {
	"Interface\\Addons\\DXE\\Sounds\\%s\\1.ogg",
	"Interface\\Addons\\DXE\\Sounds\\%s\\2.ogg",
	"Interface\\Addons\\DXE\\Sounds\\%s\\3.ogg",
	"Interface\\Addons\\DXE\\Sounds\\%s\\4.ogg",
	"Interface\\Addons\\DXE\\Sounds\\%s\\5.ogg",
	"Interface\\Addons\\DXE\\Sounds\\%s\\6.ogg",
	"Interface\\Addons\\DXE\\Sounds\\%s\\7.ogg",
	"Interface\\Addons\\DXE\\Sounds\\%s\\8.ogg",
	"Interface\\Addons\\DXE\\Sounds\\%s\\9.ogg",
	"Interface\\Addons\\DXE\\Sounds\\%s\\10.ogg",
}
function prototype:PlayAudioCountdown(sound)
	Sound(sound)
end
function prototype:AudioCountdown(totaltime)
	local voice = gbl.profile.Sounds.VoiceSound

	for i,sound in ipairs(CDList) do
		if totaltime - i == 0 then break end
		local sound2 = format(sound, tostring(voice))
		self:ScheduleTimer("PlayAudioCountdown",totaltime - i,sound2)
	end		
end

function prototype:CancelAudioCountdown()
	self:CancelAllTimers()
end

---------------------------------------
-- API
---------------------------------------

function module:QuashAll()
	for bar in pairs(Active) do bar:Destroy() end
end

function module:QuashByPattern(pattern)
	for bar in pairs(Active) do
		if bar:GetID() and find(bar:GetID(),pattern) then
				bar:Destroy()
		end
	end
end

function module:SetTimeleft(id,time)
	for bar in pairs(Active) do
		if bar:GetID() == id then
			bar.data.endTime = GetTime() + time
		end
	end
end

function module:GetTimeleft(id)
	for bar in pairs(Active) do
		if bar:GetID() == id then
			return bar.data.timeleft
		end
	end
	return -1
end

function module:Resumebar(id)
	for bar in pairs(Active) do
		if bar:GetID() == id then
			bar.paused = false
			bar:Countdown(bar.data.timeleft,bar.flashTime)
			if bar.flashTime then
				local waitTime = bar.data.timeleft - bar.flashTime
				if waitTime < 0 then bar:TranslateToCenter()
				else bar:ScheduleTimer("TranslateToCenter",waitTime - 1) end
			end
			bar:ScheduleTimer("Fade",bar.data.timeleft)
		end
	end
end

function module:Pausebar(id)
	for bar in pairs(Active) do
		if bar:GetID() == id then
			--bar.countFunc = nil
			bar.paused = true
			bar:CancelTimer("Fade")
		end
	end
end

function module:IsActive(id)
	for bar in pairs(Active) do
		if bar:GetID() == id then
			return true
		end
	end
end

function module:Auras(id, text, totalTime, flashTime, sound, c1, c2, flashscreen, icon)
	if c1 == nil or c1 == "NEWBLUE" then c1 = "MIDBLUE" end
	if pfl.DisableDropdowns then self:CenterPopup(id, text, totalTime, flashTime, sound, c1, c2, flashscreen, icon, audiocd) return end
	local soundFile,c1Data,c2Data = GetMedia(sound,c1,c2)
	self:QuashByPattern(id)
	local bar = GetBar()
	bar:SetID(id)
	bar:SetIcon(icon)
	bar:SetTimeleft(totalTime)
	bar:SetText(text)
	bar:SetFlashScreen(flashscreen)
	bar:SetColor(c1Data,c2Data)
	--bar:SetSound(soundFile)
	bar.totalTime = totalTime
	bar.flashTime = flashTime
	bar:Countdown(totalTime,flashTime)

	bar:AnchorToAuras()
	bar:SetWidth(pfl.AurasBarWidth)
	bar:ScheduleTimer("Fade",totalTime)
end

function module:Dropdown(id, text, totalTime, flashTime, sound, c1, c2, flashscreen, icon, audiocd, audiotime, biggerbar, static)
	if c1 == nil or c1 == "NEWBLUE" then c1 = "MIDBLUE" end
	if pfl.DisableDropdowns then self:CenterPopup(id, text, totalTime, flashTime, sound, c1, c2, flashscreen, icon, audiocd) return end
	local soundFile,c1Data,c2Data = GetMedia(sound,c1,c2)
	local bar = GetBar()
	bar:SetID(id)
	bar:SetIcon(icon)
	if not static then bar:SetTimeleft(totalTime) end
	bar:SetText(text)
	bar:SetFlashScreen(flashscreen)
	bar:SetColor(c1Data,c2Data)
	bar:SetSound(soundFile)
	bar.totalTime = totalTime
	bar.flashTime = flashTime
	if not static then  
		bar:Countdown(totalTime,flashTime)
		if flashTime then
			local waitTime = totalTime - flashTime
			if waitTime < 0 then bar:ScheduleTimer("TranslateToCenter",0.05)
			else bar:ScheduleTimer("TranslateToCenter",waitTime - 1) end
		end
	end
	bar:AnchorToTop()
	if biggerbar then
		bar:SetWidth(pfl.TopBarWidth + 90)
	else
		bar:SetWidth(pfl.TopBarWidth)
	end
	if not static then 
		bar:ScheduleTimer("Fade",totalTime)
	else
		if c1Data then
			bar:SetColor(c1Data)
			bar.statusbar:SetValue(1)
		end
		bar.timer:Hide()
	end
	if pfl.WarningMessages then
		local popup,before = GetMessageEra(id)
		if popup then Pour(text.." - "..MMSS(totalTime),icon,c1Data) end
		if before then bar:FireBeforeMsg() end
	end

	if audiocd then
		if not audiotime then audiotime = 5 end
		local waitTimea = totalTime - audiotime - 1
		bar:ScheduleTimer("AudioCountdown",waitTimea,audiotime + 1)
	end
end

function module:CenterPopup(id, text, totalTime, flashTime, sound, c1, c2, flashscreen, icon, audiocd, audiotime, biggerbar, static)
	if c1 == nil or c1 == "NEWBLUE" then c1 = "MIDBLUE" end
	local soundFile,c1Data,c2Data = GetMedia(sound,c1,c2)
	local bar = GetBar()
	bar:SetID(id)
	bar:SetIcon(icon)
	--self.animFunc = AnimationFunc
	--if c2Data and not flashTime then flashTime = 5 end
	if not static then 
		bar:SetTimeleft(totalTime) 
		bar:Countdown(totalTime, flashTime)
		bar:SetColor(c1Data,c2Data)
	end
	bar:SetText(text)
	--bar:Countdown(totalTime, flashTime)
	bar:SetSound(soundFile)
	if pfl.WarningAnchor and pfl.RedirectCenter and totalTime <= pfl.RedirectThreshold and not static then
		bar:AnchorToWarning()
	else
		bar:AnchorToCenter()
	end
	if not static then 
		bar:ScheduleTimer("Fade",totalTime)
	else
		if c1Data and c1 ~= "NONE" then
			bar:SetColor(c1Data)
			bar.statusbar:SetValue(1)
		end
		bar.timer:Hide()
	end
	if biggerbar then
		bar:SetWidth(pfl.CenterBarWidth + 90)
	else
		bar:SetWidth(pfl.CenterBarWidth)
	end	
	if flashscreen then self:FlashScreen(c1Data) end
	if pfl.WarningMessages then
		local popup,before = GetMessageEra(id)
		if popup then Pour(text.." - "..MMSS(totalTime),icon,c1Data) end
		if before then bar:FireBeforeMsg() end
	end
	if audiocd then
		if not audiotime then audiotime = 5 end
		local waitTimea = totalTime - audiotime - 1
		bar:ScheduleTimer("AudioCountdown",waitTimea,audiotime + 1)
	end
end
function module:Simple(text, totalTime, sound, c1, flashscreen, icon, textsize)
	if c1 == nil or c1 == "NEWBLUE" then c1 = "MIDBLUE" end
	local soundFile,c1Data = GetMedia(sound,c1)
	if soundFile and not pfl.DisableSounds then Sound(soundFile) end
	if flashscreen then self:FlashScreen(c1Data) end

	if pfl.EnableWarningMessage then
		addon:NewWarning(text,c1Data,totalTime,icon, textsize)
	else
		if pfl.WarningBars then
			local bar = GetBar()
			if c1Data then
				bar:SetColor(c1Data)
				bar.statusbar:SetValue(1)
			end
			bar:SetIcon(icon)
			bar:SetText(text)
			bar.timer:Hide()
			bar[pfl.WarningAnchor and "AnchorToWarning" or "AnchorToCenter"](bar)
			bar:ScheduleTimer("Fade",totalTime)
		end
	end
	
	if pfl.WarningMessages and pfl.WarnPopupMessage then Pour(text,icon,c1Data) end
end
function addon:BossDefeated(text, totalTime, sound, c1, flashscreen, icon, textsize)
	if c1 == nil or c1 == "NEWBLUE" then c1 = "MIDBLUE" end
	module:Simple(text, totalTime, sound, c1, flashscreen, icon, 32)
end

function module:DebuffPopup(id, text, totalTime, flashTime, sound, c1, c2, flashscreen, icon, audiocd, audiotime, biggerbar, static)
	if pfl.DebuffBars then
		if c1 == nil or c1 == "NEWBLUE" then c1 = "MIDBLUE" end
		local soundFile,c1Data,c2Data = GetMedia(sound,c1,c2)
		local bar = GetBar()
		bar:SetID(id)
		bar:SetIcon(icon)
		if not static then bar:SetTimeleft(totalTime) end
		if c1Data and c1 ~= "NONE" then
			bar:SetColor(c1Data)
			bar.statusbar:SetValue(1)
		end
		bar:SetText(text)
		bar:SetSound(soundFile)
		bar["AnchorToDebuff"](bar, biggerbar)
		--if totalTime > 0 then 
	--		bar:SetTimeleft(totalTime)
		if not static then
			bar:Countdown(totalTime, flashTime)
			bar:ScheduleTimer("Fade",totalTime)
		else
			bar.timer:Hide()
		end
		if biggerbar then
			bar:SetWidth(pfl.DebuffBarWidth + 90)
		else
			bar:SetWidth(pfl.DebuffBarWidth)
		end
		if flashscreen then self:FlashScreen(c1Data) end
		if pfl.WarningMessages then
			local popup,before = GetMessageEra(id)
			if popup then Pour(text.." - "..MMSS(totalTime),icon,c1Data) end
			if before then bar:FireBeforeMsg() end
		end
	end
	if pfl.WarningMessages and pfl.WarnPopupMessage then Pour(text,icon,c1Data) end
end
function module:Inform(text,sound, c1,totalTime,flashscreen,icon,textsize)
	if c1 == nil or c1 == "NEWBLUE" then c1 = "MIDBLUE" end
	local soundFile,c1Data = GetMedia(sound,c1)
	if soundFile and not pfl.DisableSounds then Sound(soundFile) end
	if flashscreen then self:FlashScreen(c1Data) end
	addon:InformWarning(text,c1Data,totalTime,icon,textsize)
end

---------------------------------------------
-- ABSORB BARS
---------------------------------------------

local function abbrev(value)
	return value > 1000000 and ("%.2fm"):format(value / 1000000) or ("%dk"):format(value / 1000)
end

local dmg2_types = { SPELL_MISSED = true, SPELL_PERIODIC_MISSED = true, RANGE_MISSED = true }
local function Absorb_OnEvent(self,_,_,eventtype,_,_,_,_,_,dstGUID,_,_,_,misstype,dmg,_,misstype2,dmg2)
	local data = self.data
	if (misstype == "ABSORB" or misstype2 == "ABSORB") and NID[dstGUID] == data.npcid then
		local flag
		if dmg2_types[eventtype] then
			data.value = data.value + dmg2
			flag = true
		elseif eventtype == "SWING_MISSED" then
			data.value = data.value + dmg
			flag = true
		end
		if flag then
			-- reverse
			local perc = data.value / data.total
			if perc <= 0 or perc > 1 then self:Destroy() return end
			self:SetText(data.textformat:format(abbrev(data.total - data.value),data.atotal,(1-perc) * 100))
			self.statusbar:SetValue(1 - perc)
		end
	end
end

function prototype:SetTotal(total) self.data.total,self.data.atotal = total,abbrev(total) end
function prototype:SetNPCID(npcid) self.data.npcid = npcid end
function prototype:SetUnitName(unit) self.data.unit = unit end
function prototype:SetTextFormat(textformat) self.data.textformat = textformat end
function prototype:UnregisterCLEU() self:UnregisterEvent("COMBAT_LOG_EVENT_UNFILTERED") end

function module:Absorb(id, text, textFormat, totalTime, flashTime, sound, c1, c2, flashscreen, icon, total, npcid)
	if c1 == nil or c1 == "NEWBLUE" then c1 = "MIDBLUE" end
	local soundFile,c1Data,c2Data = GetMedia(sound,c1,c2)
	if soundFile and not pfl.DisableSounds then Sound(soundFile) end
	if flashscreen then self:FlashScreen(c1Data) end
	--spellid = tonumber(spellid)
	npcid = tonumber(npcid)
	local bar = GetBar()
	bar.data.value = 0
	bar:SetID(id)
	bar:AnchorToCenter()
	bar:SetText(textFormat:format(abbrev(total),abbrev(total),100))
	bar:SetTotal(total)
	bar:SetNPCID(npcid)
	bar:Countdown(totalTime,flashTime,true)
	bar:SetIcon(icon)
	bar:SetTextFormat(textFormat)
	bar:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
	bar:SetScript("OnEvent",Absorb_OnEvent)
	bar.statusbar:SetValue(1)
	bar:ScheduleTimer("Fade",totalTime)
	bar:ScheduleTimer("UnregisterCLEU",totalTime)

	if c1Data then bar:SetColor(c1Data,c2Data) end
	if pfl.WarningMessages and pfl.WarnPopupMessage then Pour(text,icon,c1Data) end
end

--
local function UnitAbsorb_OnEvent(self,_,_,eventtype,_,_,_,_,_,dstGUID,dstName)
	local data = self.data
	
	if eventtype == "UNIT_ABSORB_AMOUNT_CHANGED" then
		data.value = UnitGetTotalAbsorbs(dstName)
		local perc = data.value / data.total
		if perc <= 0 or perc > 1 then self:Destroy() return end
		self:SetText(data.textformat:format(abbrev(data.total - data.value),data.atotal,(1-perc) * 100))
		self.statusbar:SetValue(1 - perc)
	end
end
function module:UnitAbsorb(id, text, textFormat, totalTime, flashTime, sound, c1, c2, flashscreen, icon, spellid, unit)
	if c1 == nil or c1 == "NEWBLUE" then c1 = "MIDBLUE" end
	local soundFile,c1Data,c2Data = GetMedia(sound,c1,c2)
	if soundFile and not pfl.DisableSounds then Sound(soundFile) end
	if flashscreen then self:FlashScreen(c1Data) end
	--spellid = tonumber(spellid) -- needed ?
	local bar = GetBar()
	bar.data.value = 0
	bar:SetID(id)
	bar:AnchorToCenter()
	local total = UnitGetTotalAbsorbs(unit)
	bar:SetText(textFormat:format(abbrev(total),abbrev(total),100))
	bar:SetTotal(total)
	bar:SetUnitName(unit)
	bar:Countdown(totalTime,flashTime,true)
	bar:SetIcon(icon)
	bar:SetTextFormat(textFormat)
	bar:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
	bar:SetScript("OnEvent",UnitAbsorb_OnEvent)
	bar.statusbar:SetValue(1)
	bar:ScheduleTimer("Fade",totalTime)
	bar:ScheduleTimer("UnregisterCLEU",totalTime)

	if c1Data then bar:SetColor(c1Data,c2Data) end
	if pfl.WarningMessages and pfl.WarnPopupMessage then Pour(text,icon,c1Data) end
end


---------------------------------------------
-- CUSTOM BARS
---------------------------------------------
local YOU_PREFIX = L["YOU"]..": "
local L_ICON = "Interface\\Icons\\INV_Misc_PocketWatch_02"

--do
	local R_ICON = "Interface\\Icons\\INV_Misc_PocketWatch_01"
	local ID_PREFIX = "dxecustom"
	local MSG_PTN = "^([%d:]+)%s+(.*)"
	local MSG_PTN2 = "^([%d:]+)(.*)"
	local TIME_PTN = "^(%d+):(%d+)$"
	local DROPDOWN_THRES = 15
	local FORMAT_ERROR = L["Invalid input. Usage: |cffffd200%s time text|r"]
	local TIME_ERROR = L["Invalid time. The format must be |cffffd200minutes:seconds|r or |cffffd200seconds|r (e.g. 1:30 or 90)"]
	local OFFICER_ERROR = L["You need to have raid assist"]
	local COMMTYPE = "AlertsRaidBar"

	local helpers = {
		function() return UnitExists("target") and UnitName("target") or "<<"..L["None"]..">>" end,
	}

	local function replace(text)
		text = gsub(text,"%%t",helpers[1]())
		return text
	end	
	
	function fire(text,time,color,icon)
		local id = ID_PREFIX--ID_PREFIX..text
		module:QuashByPattern(id)
		if time > DROPDOWN_THRES then
			module:Dropdown(id,text,time,DROPDOWN_THRES,pfl.CustomSound,color,nil,nil,icon,pfl.CustomAudioCD,pfl.CustomAudiotime,pfl.CustomBiggerBar)
		else
			module:CenterPopup(id,text,time,nil,pfl.CustomSound,color,nil,nil,icon,pfl.CustomAudioCD,pfl.CustomAudiotime,pfl.CustomBiggerBar)
		end
	end

	local function parse(msg,slash)
		if type(msg) ~= "string" then addon:Print(format(FORMAT_ERROR,slash)) return end
		local time,text = msg:match(MSG_PTN)
		if not time then
			time = msg:match(MSG_PTN2)
			if not time then
				addon:Print(format(FORMAT_ERROR,slash)) 
				return
			end
		end
		if not text then text = L.alert["Custom"] end
		local secs = tonumber(time)
		if not secs then
			local m,s = time:trim():match(TIME_PTN)
			if m then secs = (tonumber(m)*60) + tonumber(s)
			else addon:Print(TIME_ERROR) return end
		end
		return true,secs,replace(text)
	end

	local function LocalBarHandler(msg)
		local success,time,text = parse(msg,"/dxelb")
		if success then fire(YOU_PREFIX..text,time + 1,pfl.CustomLocalClr,L_ICON) end
	end

	local function RaidBarHandler(msg)
		if UnitIsGroupAssistant("player") or UnitIsGroupLeader("player") then
			local success,time,text = parse(msg,"/dxerb")
			if success then
				fire(YOU_PREFIX..text,time + 1,pfl.CustomRaidClr,R_ICON)
				addon:SendRaidComm(COMMTYPE,time,UnitName("player")..": "..text)
				--SendAddonMessage("D4", "PT\t".. time, IsInGroup(2) and "INSTANCE_CHAT" or "RAID")
				local _, _, _, _, _, _, _, mapID = GetInstanceInfo()
				SendAddonMessage("D4", ("PT\t%s\t%d"):format(time, mapID or 0), IsInGroup(2) and "INSTANCE_CHAT" or "RAID")
			end
		else
			addon:Print(OFFICER_ERROR)
		end
	end

	module["OnComm"..COMMTYPE] = function(self,event,commType,sender,time,text)
		--if not UnitIsGroupAssistant(sender) then return end --UnitIsRaidOfficer
		if UnitIsGroupAssistant("player") or UnitIsGroupLeader("player") then
			fire(text,time,pfl.CustomRaidClr,R_ICON)
		end
	end

	addon.RegisterCallback(module,"OnComm"..COMMTYPE)

	SlashCmdList.DXEALERTLOCALBARCUSTOM = LocalBarHandler
	SlashCmdList.DXEALERTRAIDBARCUSTOM = RaidBarHandler

	SLASH_DXEALERTLOCALBARCUSTOM1 = "/dxelb"
	SLASH_DXEALERTRAIDBARCUSTOM1 = "/dxerb"
	
	---------------------------------------------
	-- Pull BARS
	---------------------------------------------
	local ID_PREFIX2 = "dxepull"
	local DROPDOWN_THRES2 = 15
	local COMMTYPE2 = "AlertsRaidBarPull"

	function firepull(text,time,color,icon,audiotime)
		local id = ID_PREFIX2--ID_PREFIX..text
		module:QuashByPattern(id)
		if time > DROPDOWN_THRES2 then
			module:Dropdown(id,text,time,DROPDOWN_THRES2,pfl.PullSound,color,nil,nil,icon,pfl.PullAudioCD,pfl.PullAudiotime,pfl.PullBiggerBar)
		else
			module:CenterPopup(id,text,time,nil,pfl.PullSound,color,nil,nil,icon,pfl.PullAudioCD,pfl.PullAudiotime,pfl.PullBiggerBar)
		end
	end

	local function parsepull(msg,slash)
		if type(msg) ~= "string" then addon:Print(format(FORMAT_ERROR,slash)) return end
		local time,text = msg:match(MSG_PTN)
		if not time then
			time = msg:match(MSG_PTN2)
			if not time then
				addon:Print(format(FORMAT_ERROR,slash)) 
				return
			end
		end
		if not text then text = L.alert["Pulling in..."] end
		local secs = tonumber(time)
		if not secs then
			local m,s = time:trim():match(TIME_PTN)
			if m then secs = (tonumber(m)*60) + tonumber(s)
			else addon:Print(TIME_ERROR) return end
		end
		return true,secs,replace(text)
	end

	local function LocalBarHandler2(msg)
		local success,time,text = parsepull(msg,"/lpull")
		if success then firepull(YOU_PREFIX..text,time + 1,pfl.PullLocalClr,L_ICON) end
	end

	local function RaidBarHandler2(msg)
		if UnitIsGroupAssistant("player") or UnitIsGroupLeader("player") then
			local success,time,text = parsepull(msg,"/pull")
			if success then
				firepull(YOU_PREFIX..text,time + 1,pfl.PullRaidClr,R_ICON)
				local nick = UnitName("player"):gsub("%-.+", "*")
				addon:SendRaidComm(COMMTYPE2,time,nick..": "..text)
				local _, _, _, _, _, _, _, mapID = GetInstanceInfo()
				SendAddonMessage("D4", ("PT\t%s\t%d"):format(time, mapID or 0), IsInGroup(2) and "INSTANCE_CHAT" or "RAID")
			end
		else
			addon:Print(OFFICER_ERROR)
		end
	end

	module["OnComm"..COMMTYPE2] = function(self,event,commType,sender,time,text)
		if UnitIsGroupAssistant("player") or UnitIsGroupLeader("player") then
			firepull(text,time,pfl.PullRaidClr,R_ICON)
		end
	end

	addon.RegisterCallback(module,"OnComm"..COMMTYPE2)

	SlashCmdList.DXEALERTLOCALBAR = LocalBarHandler2
	SlashCmdList.DXEALERTRAIDBAR = RaidBarHandler2

	SLASH_DXEALERTLOCALBAR1 = "/lpull"
	SLASH_DXEALERTRAIDBAR1 = "/pull"
	
---------------------- Break Timer ----------------------------	
	local ID_PREFIX3 = "dxebreak"
	local DROPDOWN_THRES3 = 15
	local COMMTYPE3 = "AlertsRaidBarBreak"

	function fireBreak(text,time,color,icon)
		local id = ID_PREFIX3--ID_PREFIX..text
		module:QuashByPattern(id)
		if time > DROPDOWN_THRES3 then
			module:Dropdown(id,text,time,DROPDOWN_THRES3,pfl.BreakSound,color,nil,nil,icon,pfl.BreakAudioCD,pfl.BreakAudiotime,pfl.BreakBiggerBar)
		else
			module:CenterPopup(id,text,time,nil,pfl.BreakSound,color,nil,nil,icon,pfl.BreakAudioCD,pfl.BreakAudiotime,pfl.BreakBiggerBar)
		end
	end
	function module:BreakRelog(seconds, text)
		fireBreak(text,seconds,pfl.BreakLocalClr,L_ICON)
	end
	local function parsebreak(msg,slash)
		if type(msg) ~= "string" then addon:Print(format(FORMAT_ERROR,slash)) return end
		local time,text = msg:match(MSG_PTN)
		if not time then
			time = msg:match(MSG_PTN2)
			if not time then
				addon:Print(format(FORMAT_ERROR,slash)) 
				return
			end
		end
		if not text then text = L.alert["Break"] end
		local secs = tonumber(time)
		if not secs then
			local m,s = time:trim():match(TIME_PTN)
			if m then secs = (tonumber(m)*60) + tonumber(s)
			else addon:Print(TIME_ERROR) return end
		end
		return true,secs,replace(text)
	end

	local function LocalBarHandler3(msg)
		local success,seconds,text = parsebreak(msg,"/lbreak")
		if success then fireBreak(YOU_PREFIX..text,seconds + 1,pfl.BreakLocalClr,L_ICON) end
		addon.Loader:LoaderBreak(seconds, YOU_PREFIX..text)
	end

	local function RaidBarHandler3(msg)
		if UnitIsGroupAssistant("player") or UnitIsGroupLeader("player") then
			local success,time,text = parsebreak(msg,"/break")
			if success then
				fireBreak(YOU_PREFIX..text,time + 1,pfl.BreakRaidClr,R_ICON)
				addon:SendRaidComm(COMMTYPE3,time,UnitName("player")..": "..text)
				--SendAddonMessage("D4", "BT\t".. time, IsInGroup(2) and "INSTANCE_CHAT" or "RAID")
				local _, _, _, _, _, _, _, mapID = GetInstanceInfo()
				--SendAddonMessage("D4", ("BT\t%s\t%d"):format(time, mapID or 0), IsInGroup(2) and "INSTANCE_CHAT" or "RAID")
				SendAddonMessage("D4", ("BT\t%d"):format(time), IsInGroup(2) and "INSTANCE_CHAT" or "RAID") -- DBM message
				addon.Loader:LoaderBreak(time, YOU_PREFIX..text)
			end
		else
			addon:Print(OFFICER_ERROR)
		end
	end	

	module["OnComm"..COMMTYPE3] = function(self,event,commType,sender,seconds,text)
		--if not UnitIsGroupAssistant(sender) then return end --UnitIsRaidOfficer
		if UnitIsGroupAssistant(sender) or UnitIsGroupLeader(sender) then
			fireBreak(text,seconds,pfl.BreakRaidClr,R_ICON)
			addon.Loader:LoaderBreak(seconds, text)
		end
	end
	addon.RegisterCallback(module,"OnComm"..COMMTYPE3)
	
	SlashCmdList.DXEALERTLBREAK = LocalBarHandler3
	SlashCmdList.DXEALERTRAIDBREAK = RaidBarHandler3
	SLASH_DXEALERTLBREAK1 = "/lbreak"
	SLASH_DXEALERTRAIDBREAK1 = "/break"	
--end

---------------------------------------------
-- Sync
---------------------------------------------
	--local COMMTYPE = "AlertsRaidBar"
local Invoker = addon.Invoker
	--local function fire2(text,time,icon,color,flashtime)
	local function fire2(id,alert,opt)
		--local id = text
		--print("sync fire2",id,alert,opt)
		module:QuashByPattern(id)
		addon.Invoker:Sync(id,alert,opt)
		--Invoker:handlers.alert(alert)
		--[[if time > flashtime then
			module:Dropdown(id,text,time,flashtime,nil,color,nil,nil,icon,true)
		else
			module:CenterPopup(id,text,time,flashtime,nil,color,nil,nil,icon,true)
		end--]]
	end
	function module:SyncStuff(id,alert,opt)
		--fire2(text,time,icon,color,flashtime)
		--print("SyncStuff called")
		fire2(id,alert,opt)
		addon:SendRaidComm("OnCommAlertsUpdate",id,alert,opt)
		--addon:SendRaidComm("OnCommAlertsUpdate",text,time,icon,color,flashtime)
	end

--	module["OnCommAlertsUpdate"] = function(self,event,commType,sender,text,time,icon,color,flashtime)
--		fire2(text,time,icon,color,flashtime)
	module["OnCommAlertsUpdate"] = function(self,event,commType,sender,id,alert,opt)
		fire2(id,alert,opt)
	end
	addon.RegisterCallback(module,"OnCommAlertsUpdate")
--[[
	local function LocalBarHandlera(msg)
		--local success,time,text = parse(msg,"/dxelba")
		--if success then 
			--fire2("Touch of Sha",25,addon.ST[122224],"INDIGO",7)
			--fire2("la","CobaltMine")
			fire2(YOU_PREFIX..text,time,pfl.CustomLocalClr,L_ICON) 
		--end
	end
	SlashCmdList.DXEALERTLOCALBARA = LocalBarHandlera

	SLASH_DXEALERTLOCALBARA1 = "/dxelba"--]]
---------------------------------------------
-- ALERT TESTS
---------------------------------------------
local class_color = {}
for class,color in pairs(RAID_CLASS_COLORS) do
	class_color[class] = ("|cff%02x%02x%02x"):format(color.r * 255, color.g * 255, color.b * 255)
end
function module:BarTest()
	self:CenterPopup("alerttestdur", "Centerpopup Centerpopup ", 10, 5, "DXE ALERT1", gbl.profile.Globals.GlobalBarCrl, nil, nil, addon.ST[28374])
	self:Dropdown("alerttestcd", "Dropdown Dropdown Dropdown ", 20, 15, "DXE ALERT2", gbl.profile.Globals.GlobalBarCrl, "ORANGE", nil, addon.ST[64813])
	if not pfl.EnableWarningMessage then self:Simple("Warning Warning Warning",3,"DXE ALERT3","RED", nil, addon.ST[53351]) end
	self:DebuffPopup("alerttestdur", "Tank debuffs ", 20, 5, "DXE ALERT1", gbl.profile.Globals.GlobalBarCrl, "ORANGE", nil, addon.ST[28374])
	self:Absorb("alerttestdur", "Absorb", "Test Shield: %s",20, 5, "DXE ALERT1", gbl.profile.Globals.GlobalBarCrl, nil, nil, addon.ST[28374],40000,nil)
	--local Testname = UnitName("player")
	--local Testclass = select(2,UnitClass("player"))
	--addon:NewTMessage("Something on "..class_color[Testclass]..Testname,gbl.profile.Globals.GlobalBarCrl,addon.ST[28374],"ALERT13")
	--self:MessageShow("MessageShow", "teste", math.random(), math.random(), math.random(), math.random(), addon.ST[28374], "None")
	--self:AddCenterMessage("Something on "..UnitName("player"),1,0,1,addon.ST[28374])   -- (text, r, g, b, font, size, outline, sticky, loc, icon)
end --id, text, textFormat, totalTime, flashTime, sound, c1, c2, flashscreen, icon, total, npcid)
function module:NewAnnounce(text,color,type,sound)
	local soundFile,c1Data = GetMedia(sound,color)
	if type == "SAY" then
		SendChatMessage(text,"SAY")
	elseif type == "YELL" then
		SendChatMessage(text,"YELL")
	end
	if c1Data then self:FlashScreen(c1Data) end
	if soundFile and not pfl.DisableSounds then Sound(soundFile) end
end
function module:NewTMessage(text,color,icon,sound)
	addon:NewTMessage(text,color,icon,sound)
end
function module:StatisticsTest(barcolor, biggerbar)
	--module:QuashByPattern("invokerBestTime")
	self:Dropdown("invokerBestTime",L.alert["Fastest Kill"],30,nil,nil,barcolor,nil,nil,"Interface\\Icons\\spell_holy_borrowedtime",nil,nil,biggerbar)
end
function module:RespawnTest(barcolor, biggerbar)
	--module:QuashByPattern("respawnRespawnTime")
	self:CenterPopup("respawnRespawnTime",L.alert["Respawn"],10,nil,nil,barcolor,nil,nil,"Interface\\Icons\\spell_holy_borrowedtime",nil,nil,biggerbar)
end
function module:EnrageTest(barcolor,flashcolor,flashtime, audiocd, audiotime, biggerbar)
	--module:QuashByPattern("invoker_enrage")
	self:Dropdown("invoker_enrage",GetSpellInfo(8599),25,flashtime,nil,barcolor,flashcolor,nil,"Interface\\Icons\\Spell_Shadow_UnholyFrenzy",audiocd, audiotime, biggerbar)
end
function module:EngageTest(color,icon,sound,size,flash)
	self:Simple(L.alert["Boss"]..L.alert[" has been |cffC41F3BEngaged"],3,sound,color,flash,icon,size)
end
function module:DefeatedTest(color,icon,sound,size,flash)
	self:Simple(L.alert["Boss"]..L.alert[" has been |cffABD473Defeated"],3,nil,color,flash,icon,size)
	PlaySoundFile(SM:Fetch("sound",sound))
end
function module:PullTest()
	firepull(YOU_PREFIX..L.alert["Pulling in..."],18,pfl.PullLocalClr,L_ICON)
end
function module:BreakTest()
	fireBreak(YOU_PREFIX..L.alert["Break for Coffe"],18,pfl.BreakLocalClr,L_ICON)
end
function module:DXELBTest()
	fire(YOU_PREFIX..L.alert["Pizza Timer"],18,pfl.CustomLocalClr,L_ICON)
end
function module:GlobalTest(barcolor)
	module:QuashByPattern("alerttestdrop")
	self:Dropdown("alerttestdrop", "Dropdown to Center ", 25, 15, nil, barcolor, nil, nil, addon.ST[64813])
end
function module:AudioCountdownTest(text)
	module:QuashByPattern("alerttestdrop")
	self:Dropdown("alerttestdrop", text, 11, 10, nil, "MIDBLUE", nil, nil, addon.ST[64813],true,10)
end
local lookup
function module:FlashTest()
	if not lookup then
		lookup = {}
		for k,v in pairs(Colors) do lookup[#lookup+1] = k end
	end
	local i = math.random(1,#lookup)
	local c = Colors[lookup[i]]
	self:FlashScreen(c)
end

----------------------------------
-- COMM RECEIVED
----------------------------------
--RaidBossEmoteFrame:UnregisterAllEvents()
local timer, prevUpgradedUser = nil, nil
local function sendMsg()
	if IsInGroup() then
		SendAddonMessage("D4", "V\t"..DBMdotRevision.."\t"..DBMdotReleaseRevision.."\t"..DBMdotDisplayVersion.."\t"..GetLocale().."\t".."true", IsInGroup(2) and "INSTANCE_CHAT" or "RAID")
	end
	timer, prevUpgradedUser = nil, nil
end
function module:OnCommReceived(prefix, msg, dist, sender)
	sender = Ambiguate(sender, "none")

	if sender == addon.PNAME then
		return
	end
	--print("A",prefix, dist, sender)
	if prefix == "D4" then
		--	print(msg:gsub("\t", "__"))
		local handler, arg1, arg2, arg3, arg4 = ("\t"):split(msg)
		local time = tonumber(arg1)
		if not sender then sender = "unknown" end
		
		if handler == "PT" then
			firepull(sender..L.alert[": Pulling"],time,pfl.PullRaidClr,"Interface\\Icons\\INV_Misc_PocketWatch_01")
		elseif handler == "BT" then
			fireBreak(sender..L.alert[": Break"],time,pfl.BreakRaidClr,"Interface\\Icons\\INV_Misc_PocketWatch_01")
			addon.Loader:LoaderBreak(time, sender..L.alert[": Break"])
		elseif handler == "H" then
			--if timer then timer:Cancel() end
			--timer = CTimerNewTicker(3.3, sendMsg, 1)
		elseif handler == "V" then
			


			--[[elseif handler == "NS" then
			local popupkey = format("DXE_Confirm_%s","lala")
			local _, bossid, varname, text, abilityname = ("\t"):split(msg)
			--if not StaticPopupDialogs[popupkey] then
				local STATIC_CONFIRM = {
					text = bossid.." - "..varname.." - "..text.." - "..abilityname,
					OnAccept = function()
						self:DBMShare(bossid,sender,varname,text)
					end,
					OnCancel = function()
					--	self:DLRejected()
					end,
					button1 = L["Accept"],
					button2 = L["Reject"],
					--timeout = 15,
					whileDead = 1,
					hideOnEscape = 1,
				}
				StaticPopupDialogs[popupkey] = STATIC_CONFIRM
			--end
			StaticPopup_Show(popupkey)--]]			
		end
	end
	--[[if prefix == "BigWigs" then
		local time, text = string.match(msg, "BWCustomBar%s*(%d+)%s*(.+)")
		local handler, time2, name = ("T:"):split(msg)
		local sync, rest = msg:match("(%S+)%s*(.*)$")
		
		time = tonumber(time)
		print("Bigwigs",msg,time,text,sender,name,sync,handler,time2)
		if handler == "BWPull" then
			print("BWPull found",rest)
		end--]]
end
--[[function module:DBMShare(bossid,sender,varname,text)
	print("DBMShare",key,bossid,sender,varname,text)	
		
	--local EDBids2 = {}
	for key, data in addon:IterateEDB() do
		if data.triggers then
			local encounterid1 = data.triggers.ejid
			if tonumber(encounterid1) == tonumber(bossid) then
				print("triggers",encounterid1,key)
			end
		end
	end
end--]]

function module:GROUP_ROSTER_UPDATE()
	if IsInRaid() then
		addon.Pane.dificulty:SetDificulty(1)
	else
		if GetNumGroupMembers() > 0 then addon.Pane.dificulty:SetDificulty(2)
		else addon.Pane.dificulty:SetDificulty(0) end
	end
end


--SendAddonMessage("D4", "PT\t".. time, IsPartyLFG() and "INSTANCE_CHAT" or "RAID")
--local msg = "1111\ttestemod\tsyncText"
--self:SendMessage("M_TEST")
--SendAddonMessage("D4", "NS\t".. msg, "GUILD")
--BIGWIGS
--SendAddonMessage("BigWigs", "T:"..msg, IsPartyLFG() and "INSTANCE_CHAT" or "RAID")
