-- Interrupt Announcer
--------------------------------------------

local addon = DXE
local L = addon.L

local defaults = {
	profile = {
		Enabled = true,
		AnnounceInterrupt = false,
		AnnounceDispelled = false,
		AnnounceSpellSteal = false,
		InterruptAnnounceType = "SAY",
		InterruptChatMessages = false,
		InterruptOnlyPlayer = false,
		InterruptAllPlayers = true,
		InterruptLocalMsg = true,
		DispellAnnounceType = "SAY",
		DispellChatMessages = false,
		DispellOnlyPlayer = false,
		DispellAllPlayers = true,
		DispellLocalMsg = true,	
		SpellStealAnnounceType = "SAY",
		SpellStealChatMessages = false,
		SpellStealOnlyPlayer = false,
		SpellStealAllPlayers = true,
		SpellStealLocalMsg = true,			
	},
}

local class_color = {}
for class,color in pairs(RAID_CLASS_COLORS) do
	class_color[class] = ("|cff%02x%02x%02x"):format(color.r * 255, color.g * 255, color.b * 255)
end
-- If needed to go Global class color
local function getclasscolor(unit)
	if unit == nil then return "nil" end
	local classc = select(2,UnitClass(unit))
	if classc == nil and unit ~= nil then return unit end
	return class_color[classc]..unit.."|r" or unit
end
----------------------------------
-- INITIALIZATION
----------------------------------

local module = addon:NewModule("InterruptAnnouncer","AceEvent-3.0", "AceTimer-3.0")
addon.InterruptAnnouncer = module

local db,pfl

function module:RefreshProfile() 
	pfl = db.profile
	if not pfl.Enabled then
		--module:UnregisterEvent("GROUP_ROSTER_UPDATE")
		module:UnregisterEvent("ZONE_CHANGED_NEW_AREA")
		module:UnregisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
	elseif pfl.Enabled then
		--module:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
		--module:RegisterEvent("GROUP_ROSTER_UPDATE")
		module:RegisterEvent("ZONE_CHANGED_NEW_AREA")
		if IsInInstance() then
			self:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
		end	
	end
end

function module:OnInitialize()
	self.db = addon.db:RegisterNamespace("InterruptAnnouncer", defaults)
	db = self.db
	pfl = db.profile

	db.RegisterCallback(self, "OnProfileChanged", "RefreshProfile")
	db.RegisterCallback(self, "OnProfileCopied", "RefreshProfile")
	db.RegisterCallback(self, "OnProfileReset", "RefreshProfile")

	if pfl.Enabled then
		--self:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
		--self:RegisterEvent("GROUP_ROSTER_UPDATE")
		--self:GROUP_ROSTER_UPDATE()
		module:RegisterEvent("ZONE_CHANGED_NEW_AREA")
		if IsInInstance() then
			self:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
		end		
	end
end
function module:AnnounceType(message)
	local channel
	
	if IsInRaid() and (UnitIsGroupLeader("player") or UnitIsGroupAssistant("player")) then
		channel = "RAID_WARNING"
	elseif IsInGroup(LE_PARTY_CATEGORY_INSTANCE) then
		channel = "INSTANCE_CHAT"
	elseif IsInRaid() then
		channel = "RAID"
	elseif IsInGroup() then
		channel = "PARTY"
	else
		return
	end

	SendChatMessage("DXE: "..message, channel)
end

local throttles = {}

function module:COMBAT_LOG_EVENT_UNFILTERED(a, b, event, c, d, source, e, f, g, destName, h, i, spellId, spellName,_,_,extraSpellID)
		if source and spellName then
		
			-- Throttling
			local var = event..source
			throttles[var] = throttles[var] or 0
			-- Check throttle
			local t = GetTime()
			if t - throttles[var] > 1.5 then
					throttles[var] = t
			else
				-- Failed throttle, exit out
				return true
			end

			local message
			if event == "SPELL_INTERRUPT" then
				if source == UnitName("player") then
					if pfl.AnnounceInterrupt then
						message = format("%s %s %s %s %s",source,L["Interrupted"],extraSpellID,L.alert["on"],destName)
						SendChatMessage(message,pfl.InterruptAnnounceType)
					else 
						if pfl.InterruptChatMessages then
							if not pfl.InterruptLocalMsg then
								message = format("%s %s %s %s",L["I have interrupted"],extraSpellID,L.alert["on"],destName)
								self:AnnounceType(message)
							else
								destName = getclasscolor(destName)
								message = format("%s|r|h %s %s %s",L["I have interrupted"],extraSpellID,L.alert["on"],destName:gsub("%-.+", "*"))
								addon:Print(message) 
							end
						end	
					end
				else
					if pfl.InterruptChatMessages then
						if pfl.InterruptAllPlayers then
							if not pfl.InterruptLocalMsg then
								message = format("%s %s %s %s %s",source:gsub("%-.+", "*"),L["has interrupted"],extraSpellID,L.alert["on"],destName)
								self:AnnounceType(message)
							else
								source = getclasscolor(source)
								destName = getclasscolor(destName)
								message = format("%s|r|h %s %s %s %s",source:gsub("%-.+", "*"),L["has interrupted"],extraSpellID,L.alert["on"],destName:gsub("%-.+", "*"))
								addon:Print(message) 
							end					
						end
					end			
				end
			end
			if event == "SPELL_DISPEL" then
				if source == UnitName("player") then
					if pfl.AnnounceDispelled then
						message = format("%s %s %s",source,L["Dispelled"],extraSpellID,L.alert["on"],destName)
						SendChatMessage(message,pfl.DispellAnnounceType)
					else 
						if pfl.DispellChatMessages then
							if not pfl.DispellLocalMsg then
								message = format("%s %s %s %s",L["I have dispelled"],extraSpellID,L.alert["on"],destName)
								self:AnnounceType(message)
							else
								destName = getclasscolor(destName)
								message = format("%s %s %s %s",L["I have dispelled"],extraSpellID,L.alert["on"],destName:gsub("%-.+", "*"))
								addon:Print(message) 
							end
						end	
					end
				else
					if pfl.DispellAllPlayers then
						if not pfl.DispellLocalMsg then
							message = format("%s %s %s %s %s",source:gsub("%-.+", "*"),L["has dispelled"],extraSpellID,L.alert["on"],destName)
							self:AnnounceType(message)
						else
							source = getclasscolor(source)
							destName = getclasscolor(destName)
							message = format("%s|r|h %s %s %s %s",source:gsub("%-.+", "*"),L["has dispelled"],extraSpellID,L.alert["on"],destName:gsub("%-.+", "*"))
							addon:Print(message) 
						end
					end			
				end			
			end
			if event == "SPELL_STOLEN" then
				if source == UnitName("player") then
					if pfl.AnnounceSpellSteal then
						message = format("%s %s %s",source,L["Spellstole"],extraSpellID,L.alert["on"],destName)
						SendChatMessage(message,pfl.DispellAnnounceType)
					else 
						if pfl.SpellStealChatMessages then
							if not pfl.SpellStealAllPlayers then
								message = format("%s %s %s %s",L["I have spellstole"],extraSpellID,L.alert["on"],destName)
								self:AnnounceType(message)
							else
								destName = getclasscolor(destName)
								message = format("%s %s %s %s",L["I have spellstole"],extraSpellID,L.alert["on"],destName:gsub("%-.+", "*"))
								addon:Print(message) 
							end
						end	
					end
				else
					if pfl.SpellStealChatMessages then
						if not pfl.SpellStealAllPlayers then 
							message = format("%s %s %s %s %s",source:gsub("%-.+", "*"),L["has spellstole"],extraSpellID,L.alert["on"],destName)
							self:AnnounceType(message)
						else
							source = getclasscolor(source)
							destName = getclasscolor(destName)
							message = format("%s|r|h %s %s %s %s",source:gsub("%-.+", "*"),L["has spellstole"],extraSpellID,L.alert["on"],destName:gsub("%-.+", "*"))
							addon:Print(message) 
						end
					end			
				end			
			end			
		end
end
function module:ZONE_CHANGED_NEW_AREA()
	if not IsInInstance() or IsInInstance() == "none" then
		--instance = false
		self:UnregisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
	else
	--	instance = true
		self:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
	end
end